from pathlib import Path

from sqlalchemy.orm import Session
from fastapi.testclient import TestClient

from labelu.internal.common.config import settings
from labelu.internal.domain.models.task import Task
from labelu.internal.adapter.persistence import crud_user
from labelu.internal.adapter.persistence import crud_task
from labelu.internal.adapter.persistence import crud_attachment


class TestClassTaskAttachmentRouter:
    def test_upload_file_successful(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]

        # run
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            png_res = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": ("test.png", f, "image/png")},
            )
        # check
        json = png_res.json()
        assert png_res.status_code == 201
        
        parts = json["data"]["url"].split("/")[-3:]
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()
        parts[-1] = parts[-1][:8] + "-test-thumbnail.png"
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()
    
        with Path("labelu/tests/data/test.jsonl").open(mode="rb") as f:
            jsonl_res = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": ("test.jsonl", f, "application/json")},
            )
            
        json = jsonl_res.json()
        assert jsonl_res.status_code == 201
        
        parts = json["data"]["url"].split("/")[-3:]
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()

    def test_upload_file_when_task_finished(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                config="config",
                media_type="IMAGE",
                status="FINISHED",
            ),
        )

        # run
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            new_res = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )

        # check
        assert new_res.status_code == 201

    def test_upload_file_not_image(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]

        # run
        with Path("labelu/tests/data/test.txt").open(mode="rb") as f:
            new_res = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": ("test.txt", f, "plain/text")},
            )

        # check
        json = new_res.json()
        assert new_res.status_code == 201

        parts = json["data"]["url"].split("/")[-3:]
        assert Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()
        parts[-1] = parts[-1][:8] + "-test-thumbnail.txt"
        assert not Path(f"{settings.MEDIA_ROOT}").joinpath("/".join(parts)).exists()

    def test_upload_file_when_task_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data

        # run
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            new_res = client.post(
                f"{settings.API_V1_STR}/tasks/0/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )

        # check
        assert new_res.status_code == 404
        assert new_res.json()["err_code"] == 50002

    def test_download_file_successful(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]

        # upload file
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            attachment = client.post(
                f"{settings.API_V1_STR}/tasks/{task_id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        url = attachment.json()["data"]["url"]

        # run
        r = client.get(
            url,
            headers=testuser_token_headers,
        )

        # check
        assert r.status_code == 200

    def test_download_file_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )
        task_id = task.json()["data"]["id"]

        # run
        r = client.get(
            url=f"{settings.SCHEME}://{settings.HOST}:{settings.PORT}{settings.API_V1_STR}/tasks/attachment/upload/1/0",
            headers=testuser_token_headers,
        )

        # check
        assert r.status_code == 404
        assert r.json()["err_code"] == 51001

    def test_task_delete_task_file(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=current_user.id,
                updated_by=current_user.id,
            ),
        )
        # upload file
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            attachment = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        attachment_id = attachment.json()["data"]["id"]
        attachment_path = attachment.json()["data"]["url"]

        # run
        data = {"attachment_ids": [attachment_id]}
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        deleted = crud_attachment.get(db=db, attachment_id=attachment_id)
        file_full_path = Path(settings.MEDIA_ROOT).joinpath(attachment_path)
        assert r.status_code == 200
        assert not deleted
        assert not file_full_path.exists()

    def test_task_delete_task_file_not_found_task(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data

        # run
        data = {"attachment_ids": [1]}
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/0/attachments",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        assert r.status_code == 404
        assert r.json()["err_code"] == 51001

    def test_task_delete_task_file_not_owner(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )

        # run
        data = {"attachment_ids": [1]}
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        assert r.status_code == 403
        assert r.json()["err_code"] == 30001



from datetime import timedelta

from loguru import logger
from fastapi import status
from sqlalchemy.orm import Session

from labelu.internal.domain.models.user import User
from labelu.internal.adapter.persistence import crud_user
from labelu.internal.common.config import settings
from labelu.internal.common.error_code import ErrorCode
from labelu.internal.common.error_code import LabelUException
from labelu.internal.common.security import AccessToken
from labelu.internal.common.security import verify_password
from labelu.internal.common.security import get_password_hash
from labelu.internal.common.security import create_access_token
from labelu.internal.application.command.user import LoginCommand
from labelu.internal.application.command.user import SignupCommand
from labelu.internal.application.response.user import SignupResponse
from labelu.internal.application.response.user import LoginResponse


async def signup(db: Session, cmd: SignupCommand) -> SignupResponse:
    # check user alredy exists
    user = crud_user.get_user_by_username(db, username=cmd.username)
    if user:
        logger.error("user already exist:{}", cmd.username)
        raise LabelUException(
            code=ErrorCode.CODE_40001_USERNAME_ALREADY_EXISTS,
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    # new a user
    with db.begin():
        user = crud_user.create(
            db,
            User(
                username=cmd.username,
                hashed_password=get_password_hash(cmd.password),
            ),
        )

    # response
    return SignupResponse(id=user.id, username=user.username)


async def login(db: Session, cmd: LoginCommand) -> LoginResponse:

    # check user exsit and verify password
    user = crud_user.get_user_by_username(db, cmd.username)
    if not user or not verify_password(cmd.password, user.hashed_password):
        raise LabelUException(
            code=ErrorCode.CODE_40000_USERNAME_OR_PASSWORD_INCORRECT,
            status_code=status.HTTP_401_UNAUTHORIZED,
        )

    # create access token
    access_token_expires = timedelta(minutes=settings.TOKEN_ACCESS_EXPIRE_MINUTES)
    access_token = create_access_token(
        token=AccessToken(id=user.id, username=user.username),
        expires_delta=access_token_expires,
    )

    # response
    return LoginResponse(token=f"{settings.TOKEN_TYPE} {access_token}")



"""add file and pre_annotation

Revision ID: bc8fcb35b66b
Revises: 1b174ca5159a
Create Date: 2024-02-07 15:58:30.618151

"""
import imp
import json
import os

from alembic import context, op
import sqlalchemy as sa
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import sessionmaker
from labelu.internal.common.config import settings

Base = automap_base()


# revision identifiers, used by Alembic.
revision = 'bc8fcb35b66b'
down_revision = '1b174ca5159a'
branch_labels = None
depends_on = None

# import alembic_labelu_tools from the absolute path
alembic_labelu_tools = imp.load_source(
    "alembic_labelu_tools",
    (
        os.path.join(
            os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
            "alembic_labelu_tools.py",
        )
    ),
)

def upgrade() -> None:
    bind = op.get_bind()
    Base.prepare(autoload_with=bind, reflect=True)
    task_sample = Base.classes.task_sample
    # make a session
    Session = sessionmaker(bind=bind)
    session = Session()
    
    try:
    
        with context.begin_transaction():
            # Create a new table task_pre_annotation
            if not alembic_labelu_tools.table_exist("task_pre_annotation"):
                op.create_table(
                    "task_pre_annotation",
                    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True, index=True),
                    sa.Column("task_id", sa.Integer, sa.ForeignKey("task.id"), index=True),
                    sa.Column("file_id", sa.Integer, sa.ForeignKey("task_attachment.id"), index=True),
                    # json 字符串
                    sa.Column("data", sa.Text, comment="task sample pre annotation result"),
                    sa.Column("created_by", sa.Integer, sa.ForeignKey("user.id"), index=True),
                    sa.Column("updated_by", sa.Integer, sa.ForeignKey("user.id")),
                    sa.Column(
                        "created_at",
                        sa.DateTime,
                        default=sa.func.now(),
                        comment="Time a task sample result was created",
                    ),
                    sa.Column(
                        "updated_at",
                        sa.DateTime,
                        default=sa.func.now(),
                        onupdate=sa.func.now(),
                        comment="Last time a task sample result was updated",
                    ),
                    sa.Column(
                        "deleted_at",
                        sa.DateTime,
                        index=True,
                        comment="Task delete time",
                    ),
                )
            # Update the task_sample table
            if not alembic_labelu_tools.column_exist_in_table(
                "task_sample", "file_id"
            ):
                with op.batch_alter_table('task_sample', recreate="always") as batch_op:
                    batch_op.add_column(
                        sa.Column(
                            "file_id",
                            sa.Integer(),
                            sa.ForeignKey("task_attachment.id", name="fk_file_id"),
                            index=True,
                            comment="file id",
                        ),
                    )
            
            # Update the task_attachment table
            if not alembic_labelu_tools.column_exist_in_table("task_attachment", "filename"):
                with op.batch_alter_table("task_attachment", recreate="always") as batch_op_task_attachment:
                    batch_op_task_attachment.add_column(
                        sa.Column(
                            "filename",
                            sa.String(256),
                            comment="file name",
                        ),
                    )
                    batch_op_task_attachment.add_column(
                        sa.Column(
                            "url",
                            sa.String(256),
                            comment="file url",
                        ),
                    )
                            
            # Update existing data in the task_sample table
            task_items = session.execute(
                'SELECT id, config FROM task'
            )
            
            # Update the task_attachment table
            attachments = session.execute(
                'SELECT id, path FROM task_attachment'
            )
            
            for attachment in attachments:
                attachment_id = attachment[0]
                attachment_path = attachment[1]
                filename = os.path.basename(attachment_path)
                url = f"{settings.API_V1_STR}/tasks/attachment/{attachment_path}"
                
                if filename:
                    session.execute(
                        f"UPDATE task_attachment SET filename='{filename}', url='{url}' WHERE id={attachment_id}"
                    )
            
            for task_item in task_items:
                task_id = task_item[0]
                task_samples = session.execute(
                    f"SELECT id, task_attachment_ids FROM task_sample WHERE task_id={task_id}"
                )

                for task_sample in task_samples:
                    task_sample_id = task_sample[0]
                    attachment_ids = json.loads(task_sample[1])
                    # attachment_ids 存储的是字符串[id1, id2, id3]，需要转换成数组
                    file_id = attachment_ids[0]
                    
                    if not file_id:
                        continue
                    
                    attachment = session.execute(
                        f"SELECT id, path FROM task_attachment WHERE id={file_id}"
                    )
                    attachment_path = list(attachment)[0][1]
                    
                    if attachment_path:
                        # Update the task_sample table
                        session.execute(
                            f"UPDATE task_sample SET file_id={file_id} WHERE id={task_sample_id}"
                        )
            
            session.commit()
    
    except Exception as e:
        session.rollback()
        raise e
    finally:
        session.close()

def downgrade() -> None:
    op.drop_table("task_pre_annotation")
    with op.batch_alter_table('task_sample') as batch_op:
        batch_op.drop_column('file_id')



from typing import List

from sqlalchemy.orm import Session
from fastapi import APIRouter, status, Depends, Security
from fastapi.security import HTTPAuthorizationCredentials

from labelu.internal.common import db
from labelu.internal.common.security import security
from labelu.internal.domain.models.user import User
from labelu.internal.dependencies.user import get_current_user
from labelu.internal.application.service import task as service
from labelu.internal.application.command.task import BasicConfigCommand
from labelu.internal.application.command.task import UpdateCommand
from labelu.internal.application.response.base import OkResp
from labelu.internal.application.response.base import MetaData
from labelu.internal.application.response.base import CommonDataResp
from labelu.internal.application.response.base import OkRespWithMeta
from labelu.internal.application.response.task import TaskResponse
from labelu.internal.application.response.task import TaskResponseWithStatics


router = APIRouter(prefix="/tasks", tags=["tasks"])


@router.post(
    "",
    response_model=OkResp[TaskResponse],
    status_code=status.HTTP_201_CREATED,
)
async def create(
    cmd: BasicConfigCommand,
    authorization: HTTPAuthorizationCredentials = Security(security),
    db: Session = Depends(db.get_db),
    current_user: User = Depends(get_current_user),
):
    """
    Create a task with basic config.
    """

    # business logic
    data = await service.create(db=db, cmd=cmd, current_user=current_user)

    # response
    return OkResp[TaskResponse](data=data)


@router.get(
    "",
    response_model=OkRespWithMeta[List[TaskResponseWithStatics]],
    status_code=status.HTTP_200_OK,
)
async def list_by(
    page: int = 0,
    size: int = 100,
    authorization: HTTPAuthorizationCredentials = Security(security),
    db: Session = Depends(db.get_db),
    current_user: User = Depends(get_current_user),
):
    """
    List task.
    """

    # business logic
    data, total = await service.list_by(
        db=db, current_user=current_user, page=page, size=size
    )

    # response
    meta_data = MetaData(total=total, page=page, size=len(data))
    return OkRespWithMeta[List[TaskResponseWithStatics]](meta_data=meta_data, data=data)


@router.get(
    "/{task_id}",
    response_model=OkResp[TaskResponseWithStatics],
    status_code=status.HTTP_200_OK,
)
async def get(
    task_id: int,
    authorization: HTTPAuthorizationCredentials = Security(security),
    db: Session = Depends(db.get_db),
    current_user: User = Depends(get_current_user),
):
    """
    Get task detail.
    """

    # business logic
    data = await service.get(db=db, task_id=task_id, current_user=current_user)

    # response
    return OkResp[TaskResponseWithStatics](data=data)


@router.patch(
    "/{task_id}",
    response_model=OkResp[TaskResponse],
    status_code=status.HTTP_200_OK,
)
async def update(
    task_id: int,
    cmd: UpdateCommand,
    authorization: HTTPAuthorizationCredentials = Security(security),
    db: Session = Depends(db.get_db),
    current_user: User = Depends(get_current_user),
):
    """
    Update task info, inlucde annotation config, name, description and tips.
    """

    # business logic
    data = await service.update(db=db, task_id=task_id, cmd=cmd)

    # response
    return OkResp[TaskResponse](data=data)


@router.delete(
    "/{task_id}",
    response_model=OkResp[CommonDataResp],
    status_code=status.HTTP_200_OK,
)
async def delete(
    task_id: int,
    authorization: HTTPAuthorizationCredentials = Security(security),
    db: Session = Depends(db.get_db),
    current_user: User = Depends(get_current_user),
):
    """
    Delete task.
    """

    # business logic
    data = await service.delete(db=db, task_id=task_id, current_user=current_user)

    # response
    return OkResp[CommonDataResp](data=data)



import json
from zipfile import ZipFile
from PIL import Image, ImageDraw
from enum import Enum

from typing import List
from loguru import logger

from labelu.internal.common.color import colors


class Format(str, Enum):
    JSON = "JSON"
    COCO = "COCO"
    MASK = "MASK"


class Converter:
    def convert(
        self,
        config: dict,
        input_data: List[dict],
        out_data_dir: str,
        out_data_file_name_prefix: str,
        format: str,
    ) -> str:
        if format == Format.JSON.value:
            return self.convert_to_json(
                input_data=input_data,
                out_data_dir=out_data_dir,
                out_data_file_name_prefix=out_data_file_name_prefix,
            )
        elif format == Format.COCO.value:
            return self.convert_to_coco(
                config=config,
                input_data=input_data,
                out_data_dir=out_data_dir,
                out_data_file_name_prefix=out_data_file_name_prefix,
            )
        elif format == Format.MASK.value:
            return self.convert_to_mask(
                input_data=input_data,
                out_data_dir=out_data_dir,
                out_data_file_name_prefix=out_data_file_name_prefix,
            )

    def convert_to_json(
        self,
        input_data: List[dict],
        out_data_dir: str,
        out_data_file_name_prefix: str,
    ) -> str:
        out_data_dir.mkdir(parents=True, exist_ok=True)
        file_full_path = out_data_dir.joinpath("result.json")

        # every file result
        results = []
        for sample in input_data:
            data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            
            # change skipped result is invalid
            annotated_result = json.loads(data.get("result"))
            if annotated_result and sample.get("state") == "SKIPPED":
                annotated_result["valid"] = False

            # change result struct
            if annotated_result:
                annotations = []
                for tool in annotated_result.copy().keys():
                    if tool.endswith("Tool"):
                        tool_results = annotated_result.pop(tool)
                        for tool_result in tool_results.get("result", []):
                            # 视频文件的标注结果已经保存了 label 键的值，不需要再做转换
                            if "label" not in tool_result:
                                tool_result["label"] = tool_result.pop("attribute", "")

                            tool_result.pop("sourceID", None)

                            if tool == "tagTool" or tool == "textTool":
                                tool_result.pop("label")

                            if "attribute" in tool_result:
                                tool_result.pop("attribute")

                        annotations.append(tool_results)

                annotated_result["annotations"] = annotations

            annotated_result_str = json.dumps(annotated_result, ensure_ascii=False)
            results.append(
                {
                    "id": sample.get("id"),
                    "result": annotated_result_str,
                    "url": file.get("url"),
                    "fileName": file.get("filename", "")[9:],
                }
            )

        # Serializing json
        json_object = json.dumps(results, default=str)
        with file_full_path.open("w") as outfile:
            outfile.write(json_object)

        logger.info("Export file path: {}", file_full_path)
        return file_full_path

    def convert_to_coco(
        self,
        config: dict,
        input_data: List[dict],
        out_data_dir: str,
        out_data_file_name_prefix: str,
    ) -> str:

        # result output file
        out_data_dir.mkdir(parents=True, exist_ok=True)
        file_full_path = out_data_dir.joinpath("result.json")

        # result struct
        result = {
            "images": [],
            "annotations": [],
            "categories": [],
        }

        # result catetory
        category_id = 0
        logger.info("get categories")
        for attr in config.get("attributes", []):
            category = {
                "id": category_id,
                "name": attr.get("value", ""),
                "supercategory": "",
            }
            result["categories"].append(category)
            category_id += 1
        tools_category = config.get("tools", [])
        for tool in tools_category:
            for attr in tool.get("config", {}).get("attributes", []):
                category = {
                    "id": category_id,
                    "name": attr.get("value", ""),
                    "supercategory": "",
                }
                result["categories"].append(category)
                category_id += 1

        logger.info("get categories map with id")
        category_name_map_id = {}
        for category in result["categories"]:
            category_name_map_id[category.get("name")] = category.get("id")

        # annotation index
        annotation_id = 0

        # for every annotation media
        for sample in input_data:
            annotation_data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            logger.info("data is: {}", sample)

            # annotation result
            annotation_result = json.loads(annotation_data.get("result", {}))

            # coco image
            image = {
                "id": sample.get("id"),
                "fileName": file.get("filename", "")[9:],
                "width": annotation_result.get("width", 0),
                "height": annotation_result.get("height", 0),
                "valid": False
                if sample.get("state", "") == "SKIPPED"
                else annotation_result.get("valid", True),
                "rotate": annotation_result.get("rotate", 0),
            }
            result["images"].append(image)

            # every image may have multi tools
            tools = []
            if annotation_result.get("polygonTool", {}):
                tools.append(annotation_result.get("polygonTool"))
            if annotation_result.get("rectTool", {}):
                tools.append(annotation_result.get("rectTool"))

            for tool in tools:
                for tool_result in tool.get("result"):

                    bbox = []
                    segmentation = []
                    polygon_area = 0.0
                    # polygon tool
                    if tool.get("toolName") == "polygonTool":
                        x_coordinates = []
                        y_coordinates = []
                        for point in tool_result.get("points", []):
                            segmentation.append(point.get("x"))
                            segmentation.append(point.get("y"))
                            x_coordinates.append(point.get("x"))
                            y_coordinates.append(point.get("y"))

                        bbox = [
                            min(x_coordinates),
                            max(y_coordinates),
                            max(x_coordinates) - min(x_coordinates),
                            max(y_coordinates) - min(y_coordinates),
                        ]
                        polygon_area = _polygonArea(x_coordinates, y_coordinates)
                    elif tool.get("toolName") == "rectTool":
                        # rect
                        x = tool_result.get("x")
                        y = tool_result.get("y")
                        width = tool_result.get("width")
                        height = tool_result.get("height")
                        
                        if x is not None and y is not None and width is not None and height is not None:
                            bbox.extend([x, y, width, height])
                        polygon_area = tool_result.get("width", 0) * tool_result.get(
                            "height", 0
                        )

                    annotation = {
                        "image_id": sample.get("id"),
                        "id": annotation_id,
                        "iscrowd": tool_result.get("iscrowd", 0),
                        "segmentation": segmentation,
                        "area": polygon_area,
                        "bbox": bbox,
                        "category_id": category_name_map_id.get(
                            tool_result.get("label", ""), -1
                        ),
                        "order": tool_result.get("order", 0),
                    }

                    annotation_id += 1
                    result["annotations"].append(annotation)

        # Serializing json
        json_object = json.dumps(result, default=str)
        with file_full_path.open("w") as outfile:
            outfile.write(json_object)
        logger.info("Export file path: {}", file_full_path)
        return file_full_path

    def convert_to_mask(
        self,
        input_data: List[dict],
        out_data_dir: str,
        out_data_file_name_prefix: str,
    ) -> str:

        # result output file
        out_data_dir.mkdir(parents=True, exist_ok=True)

        polygon = []
        export_files = []
        color_list = []
        for sample in input_data:
            file = sample.get("file", {})
            if sample.get("state") != "DONE":
                continue
            annotation_data = json.loads(sample.get("data"))
            logger.info("data is: {}", sample)
            filename = file.get("filename")
            if filename and filename.split("/")[-1]:
                file_relative_path_base_name = filename.split("/")[-1].split(".")[0][9:]
            else:
                file_relative_path_base_name = "result"

            # annotation result
            annotation_result = json.loads(annotation_data.get("result", {}))
            if not annotation_result or not annotation_result.get("polygonTool", {}):
                continue

            # polygon tool
            polygons = []
            polygon_attribute = []
            for tool_result in annotation_result.get("polygonTool", {}).get(
                "result", []
            ):
                polygon = []
                for point in tool_result.get("points", []):
                    polygon.append(point.get("x"))
                    polygon.append(point.get("y"))
                polygons.append(polygon)
                polygon_attribute.append(tool_result.get("label", ""))

            width = annotation_result.get("width")
            height = annotation_result.get("height")

            # generate single change
            file_relative_path_model_l = f"{file_relative_path_base_name}-trainIds.png"
            file_full_path_model_l = out_data_dir.joinpath(file_relative_path_model_l)
            img_model_l = Image.new("L", (width, height), 0)
            for p in polygons:
                ImageDraw.Draw(img_model_l).polygon(p, outline=1, fill=1)
            img_model_l.save(file_full_path_model_l, "PNG")
            export_files.append(file_full_path_model_l)

            # generate RGB
            file_relative_path_model_rgb = (
                f"{file_relative_path_base_name}-segmentation.png"
            )
            file_full_path_model_rgb = out_data_dir.joinpath(
                file_relative_path_model_rgb
            )
            img_model_rgb = Image.new("RGB", (width, height), 0)
            for index, p in enumerate(polygons):
                color = colors[index % 255 + 1]
                ImageDraw.Draw(img_model_rgb).polygon(p, fill=color.get("hexString"))

                rgb = color.get("rgb")
                color_list.append(
                    {
                        "color": f'rgb({rgb.get("r")},{rgb.get("g")},{rgb.get("b")})',
                        "colorList": [rgb.get("r"), rgb.get("g"), rgb.get("b"), 255],
                        "trainIds": index + 1,
                        "attribute": polygon_attribute[index],
                    }
                )
            img_model_rgb.save(file_full_path_model_rgb, "PNG")
            export_files.append(file_full_path_model_rgb)

        # color list
        file_full_path_colors = out_data_dir.joinpath("colors.json")
        json_object = json.dumps(color_list, default=str)
        with file_full_path_colors.open("w") as outfile:
            outfile.write(json_object)
        export_files.append(file_full_path_colors)

        file_relative_path_zip = f"task-{out_data_file_name_prefix}-mask.zip"
        file_full_path_zip = out_data_dir.joinpath(file_relative_path_zip)
        with ZipFile(file_full_path_zip, "w") as zipf:
            for f in export_files:
                zipf.write(str(f), arcname=f.name)
        logger.info("Export file path: {}", file_full_path_zip)
        return file_full_path_zip


def _polygonArea(X, Y):

    # Initialize area
    area = 0.0
    n = len(X)

    # Calculate value of shoelace formula
    j = n - 1
    for i in range(0, n):
        area += (X[j] + X[i]) * (Y[j] - Y[i])
        j = i  # j is previous vertex to i

    # Return absolute value
    return abs(area) / 2.0


converter = Converter()



import json
import uuid
from datetime import datetime
from typing import List, Tuple, Union

from pathlib import Path
from loguru import logger
from fastapi import status
from sqlalchemy.orm import Session

from labelu.internal.common.config import settings
from labelu.internal.common.converter import converter
from labelu.internal.common.error_code import ErrorCode
from labelu.internal.common.error_code import LabelUException
from labelu.internal.adapter.persistence import crud_task
from labelu.internal.adapter.persistence import crud_sample
from labelu.internal.domain.models.user import User
from labelu.internal.domain.models.task import Task
from labelu.internal.domain.models.task import TaskStatus
from labelu.internal.domain.models.sample import TaskSample
from labelu.internal.domain.models.sample import SampleState
from labelu.internal.application.command.sample import ExportType
from labelu.internal.application.command.sample import PatchSampleCommand
from labelu.internal.application.command.sample import CreateSampleCommand
from labelu.internal.application.response.base import UserResp
from labelu.internal.application.response.base import CommonDataResp
from labelu.internal.application.response.sample import CreateSampleResponse
from labelu.internal.application.response.sample import SampleResponse
from labelu.internal.application.response.attachment import AttachmentResponse

async def create(
    db: Session, task_id: int, cmd: List[CreateSampleCommand], current_user: User
) -> CreateSampleResponse:
    obj_in = {}
    with db.begin():
        # check task exist
        task = crud_task.get(db=db, task_id=task_id, lock=True)
        if not task:
            logger.error("cannot find task:{}", task_id)
            raise LabelUException(
                code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
                status_code=status.HTTP_404_NOT_FOUND,
            )

        samples = [
            TaskSample(
                inner_id=task.last_sample_inner_id + i + 1,
                task_id=task_id,
                file_id=sample.file_id,
                created_by=current_user.id,
                updated_by=current_user.id,
                data=json.dumps(sample.data, ensure_ascii=False),
            )
            for i, sample in enumerate(cmd)
        ]
        obj_in[Task.last_sample_inner_id.key] = task.last_sample_inner_id + len(cmd)
        if task.status == TaskStatus.DRAFT.value:
            obj_in[Task.status.key] = TaskStatus.IMPORTED
        crud_task.update(db=db, db_obj=task, obj_in=obj_in)
        new_samples = crud_sample.batch(db=db, samples=samples)

    # response
    ids = [s.id for s in new_samples]
    return CreateSampleResponse(ids=ids)


async def list_by(
    db: Session,
    task_id: Union[int, None],
    after: Union[int, None],
    before: Union[int, None],
    pageNo: Union[int, None],
    pageSize: int,
    sorting: Union[str, None],
    current_user: User,
) -> Tuple[List[SampleResponse], int]:

    samples = crud_sample.list_by(
        db=db,
        task_id=task_id,
        owner_id=current_user.id,
        after=after,
        before=before,
        pageNo=pageNo,
        pageSize=pageSize,
        sorting=sorting,
    )

    total = crud_sample.count(db=db, task_id=task_id, owner_id=current_user.id)

    # response
    return [
        SampleResponse(
            id=sample.id,
            inner_id=sample.inner_id,
            state=sample.state,
            data=json.loads(sample.data),
            annotated_count=sample.annotated_count,
            file=AttachmentResponse(id=sample.file.id, filename=sample.file.filename, url=sample.file.url) if sample.file else None,
            created_at=sample.created_at,
            created_by=UserResp(
                id=sample.owner.id,
                username=sample.owner.username,
            ),
            updated_at=sample.updated_at,
            updated_by=UserResp(
                id=sample.updater.id,
                username=sample.updater.username,
            ),
        )
        for sample in samples
    ], total


async def get(
    db: Session, task_id: int, sample_id: int, current_user: User
) -> SampleResponse:
    sample = crud_sample.get(
        db=db,
        sample_id=sample_id,
    )

    if not sample:
        logger.error("cannot find sample:{}", sample_id)
        raise LabelUException(
            code=ErrorCode.CODE_55001_SAMPLE_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    # response
    return SampleResponse(
        id=sample.id,
        inner_id=sample.inner_id,
        state=sample.state,
        data=json.loads(sample.data),
        file=AttachmentResponse(id=sample.file.id, filename=sample.file.filename, url=sample.file.url) if sample.file else None,
        annotated_count=sample.annotated_count,
        created_at=sample.created_at,
        created_by=UserResp(
            id=sample.owner.id,
            username=sample.owner.username,
        ),
        updated_at=sample.updated_at,
        updated_by=UserResp(
            id=sample.updater.id,
            username=sample.updater.username,
        ),
    )


async def patch(
    db: Session,
    task_id: int,
    sample_id: int,
    cmd: PatchSampleCommand,
    current_user: User,
) -> SampleResponse:

    # check task exist
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    # get sample
    sample = crud_sample.get(db=db, sample_id=sample_id)
    if not sample:
        logger.error("cannot find sample:{}", sample_id)
        raise LabelUException(
            code=ErrorCode.CODE_55001_SAMPLE_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    # update
    sample_obj_in = {}
    if cmd.state == SampleState.SKIPPED.value:
        sample_obj_in[TaskSample.state.key] = SampleState.SKIPPED.value
    elif cmd.state == SampleState.NEW.value:
        sample_obj_in[TaskSample.data.key] = json.dumps(cmd.data, ensure_ascii=False)
        sample_obj_in[TaskSample.annotated_count.key] = cmd.annotated_count
        sample_obj_in[TaskSample.state.key] = SampleState.NEW.value
    else:  # can be None, or DONE
        sample_obj_in[TaskSample.data.key] = json.dumps(cmd.data, ensure_ascii=False)
        sample_obj_in[TaskSample.annotated_count.key] = cmd.annotated_count
        sample_obj_in[TaskSample.state.key] = SampleState.DONE.value

    with db.begin():
        # update task status
        if task.status != TaskStatus.FINISHED.value:
            statics = crud_sample.statics(
                db=db, owner_id=current_user.id, task_ids=[task_id]
            )
            task_obj_in = {Task.status.key: TaskStatus.INPROGRESS.value}
            new_sample_cnt = statics.get(f"{task.id}_{SampleState.NEW.value}", 0)
            if new_sample_cnt == 0 or (
                new_sample_cnt == 1 and sample.state == SampleState.NEW.value
            ):
                task_obj_in[Task.status.key] = TaskStatus.FINISHED.value
            if task.status != task_obj_in[Task.status.key]:
                crud_task.update(db=db, db_obj=task, obj_in=task_obj_in)
        # update task sample result
        updated_sample = crud_sample.update(db=db, db_obj=sample, obj_in=sample_obj_in)

    # response
    return SampleResponse(
        id=updated_sample.id,
        inner_id=updated_sample.inner_id,
        state=updated_sample.state,
        data=json.loads(updated_sample.data),
        annotated_count=updated_sample.annotated_count,
        created_at=updated_sample.created_at,
        created_by=UserResp(
            id=updated_sample.owner.id,
            username=updated_sample.owner.username,
        ),
        updated_at=updated_sample.updated_at,
        updated_by=UserResp(
            id=updated_sample.updater.id,
            username=updated_sample.updater.username,
        ),
    )


async def delete(
    db: Session, sample_ids: List[int], current_user: User
) -> CommonDataResp:

    with db.begin():
        crud_sample.delete(db=db, sample_ids=sample_ids)
    # response
    return CommonDataResp(ok=True)


async def export(
    db: Session,
    task_id: int,
    export_type: ExportType,
    sample_ids: List[int],
    current_user: User,
) -> str:

    task = crud_task.get(db=db, task_id=task_id)
    samples = crud_sample.get_by_ids(db=db, sample_ids=sample_ids)
    data = []
    for sample in samples:
       data_dict = sample.__dict__.get('data')
       if data_dict is None:
           data_dict = {}
       file_dict = sample.file.__dict__ if hasattr(sample.file, '__dict__') else {}
       data.append({ **sample.__dict__, 'file': file_dict})
       

    # output data path
    out_data_dir = Path(settings.MEDIA_ROOT).joinpath(
        settings.EXPORT_DIR,
        f"task-{task_id}-{datetime.now().strftime('%Y%m%d%H%M%S')}-{str(uuid.uuid4())[0:8]}",
    )

    # converter to export_type
    try:
        file_full_path = converter.convert(
            config=json.loads(task.config),
            input_data=data,
            out_data_dir=out_data_dir,
            out_data_file_name_prefix=task_id,
            format=export_type.value,
        )
    except Exception as e:
        logger.error(data)
        logger.error(e)
        raise LabelUException(
            code=ErrorCode.CODE_55002_SAMPLE_FORMAT_ERROR,
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )

    # response
    return file_full_path



"""update pointList => points

Revision ID: 1b174ca5159a
Revises: 0145db0fec34
Create Date: 2024-02-01 19:26:24.048019

"""
import json
from alembic import context, op
from sqlalchemy import update
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import sessionmaker

Base = automap_base()

def update_fields(sample_data: dict) -> dict:
    """update the field pointList to points in sample_data"""
    # get the result of sample_data
    annotated_result = sample_data.get("result")
    annotated_result = json.loads(annotated_result)
    # update the field pointList to points
    for sample_tool, sample_tool_results in annotated_result.items():
        if sample_tool.endswith("Tool") and isinstance(sample_tool_results, dict):
            for sample_tool_result in sample_tool_results.get("result", []):
                keys = sample_tool_result.keys()
                # replace pointList to points
                if "pointList" in keys:
                    sample_tool_result["points"] = sample_tool_result.pop("pointList")
                # replace attribute to label
                if "attribute" in keys:
                    sample_tool_result["label"] = sample_tool_result.pop("attribute")
    return annotated_result


# revision identifiers, used by Alembic.
revision = '1b174ca5159a'
down_revision = '0145db0fec34'
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Replace pointList with points in result table"""
    # get the current connection
    bind = op.get_bind()
    # Reflect ORM models from the database
    Base.prepare(autoload_with=bind, reflect=True)
    task_sample = Base.classes.task_sample
    # make a session
    Session = sessionmaker(bind=bind)
    session = Session()

    with context.begin_transaction():
        task_items = session.execute(
            'SELECT id, config FROM task WHERE config IS NOT NULL AND config != ""'
        )

        for task_item in task_items:
            task_id = task_item[0]
            
            sample_items = session.execute(
                f"SELECT id, data, annotated_count FROM task_sample WHERE task_id={task_id}"
            )
            
            for sample_item in sample_items:
                sample_id = sample_item[0]
                sample_data_item = json.loads(sample_item[1])
                if sample_item[2] > 0:
                    sample_annotated_result = update_fields(
                        sample_data_item
                    )
                    sample_data_item["result"] = json.dumps(
                        sample_annotated_result, ensure_ascii=False
                    )
                # display unicode as the corresponding Chinese
                else:
                    pass
                sample_annotated_item_str = json.dumps(
                    sample_data_item, ensure_ascii=False
                )
                op.execute(
                    update(task_sample)
                    .where(task_sample.id == sample_id)
                    .values({task_sample.data: sample_annotated_item_str})
                )

def downgrade() -> None:
    pass



import json
from datetime import datetime
from typing import Dict, List, Tuple, Optional

from pathlib import Path
from loguru import logger
from fastapi import status
from sqlalchemy.orm import Session

from labelu.internal.common.config import settings
from labelu.internal.common.error_code import ErrorCode
from labelu.internal.common.error_code import LabelUException
from labelu.internal.adapter.persistence import crud_task
from labelu.internal.adapter.persistence import crud_pre_annotation
from labelu.internal.adapter.persistence import crud_attachment
from labelu.internal.domain.models.user import User
from labelu.internal.domain.models.pre_annotation import TaskPreAnnotation
from labelu.internal.application.command.pre_annotation import CreatePreAnnotationCommand
from labelu.internal.application.response.base import UserResp
from labelu.internal.application.response.base import CommonDataResp
from labelu.internal.application.response.pre_annotation import CreatePreAnnotationResponse
from labelu.internal.application.response.pre_annotation import PreAnnotationResponse
from labelu.internal.application.response.attachment import AttachmentResponse

def read_jsonl_file(db: Session, file_id: int) -> List[dict]:
    attachment = crud_attachment.get(db, file_id)
    if attachment is None:
        raise LabelUException(status_code=404, code=ErrorCode.CODE_51001_TASK_ATTACHMENT_NOT_FOUND)

    attachment_path = attachment.path
    file_full_path = settings.MEDIA_ROOT.joinpath(attachment_path.lstrip("/"))

    try:
        with open(file_full_path, "r", encoding="utf-8") as f:
            data = f.readlines()
    except FileNotFoundError:
        raise LabelUException(status_code=404, code=ErrorCode.CODE_51001_TASK_ATTACHMENT_NOT_FOUND)

    parsed_data = [json.loads(line) for line in data]
    return parsed_data

async def create(
    db: Session, task_id: int, cmd: List[CreatePreAnnotationCommand], current_user: User
) -> CreatePreAnnotationResponse:
    with db.begin():
        # check task exist
        task = crud_task.get(db=db, task_id=task_id, lock=True)
        if not task:
            logger.error("cannot find task:{}", task_id)
            raise LabelUException(
                code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
                status_code=status.HTTP_404_NOT_FOUND,
            )

        def validate_sample_name_exists(file_id: int, pre_annotations_dict: Dict[str, List[TaskPreAnnotation]]) -> List[dict]:
            jsonl_content = read_jsonl_file(db, file_id)
            
            for item in jsonl_content:
                sample_name = item.get("sample_name")
                pre_annotations = pre_annotations_dict.get(sample_name, [])
                
                if len(pre_annotations) > 0:
                    raise LabelUException(
                        code=ErrorCode.CODE_55002_SAMPLE_NAME_EXISTS,
                        status_code=status.HTTP_400_BAD_REQUEST,
                    )
        
        # Get all pre_annotations in one query
        query_pre_annotations = crud_pre_annotation.list_by_task_id_and_owner_id(db=db, task_id=task_id, owner_id=current_user.id)
        
        pre_annotations_dict = {}
        for pre_annotation in query_pre_annotations:
            jsonl_content = read_jsonl_file(db, pre_annotation.file_id)
            
            for item in jsonl_content:
                sample_name = item.get("sample_name")

                # 如果字典中已经有这个 sample_name，就添加到列表中
                if sample_name in pre_annotations_dict:
                    pre_annotations_dict[sample_name].append(pre_annotation)
                # 否则，创建一个新的列表
                else:
                    pre_annotations_dict[sample_name] = [pre_annotation]
        
        for pre_annotation in cmd:
            validate_sample_name_exists(pre_annotation.file_id, pre_annotations_dict)
        
        pre_annotations = [
            TaskPreAnnotation(
                task_id=task_id,
                file_id=pre_annotation.file_id,
                created_by=current_user.id,
                updated_by=current_user.id,
            )
            for pre_annotation in cmd
        ]
        new_pre_annotations = crud_pre_annotation.batch(db=db, pre_annotations=pre_annotations)

    # response
    ids = [s.id for s in new_pre_annotations]
    return CreatePreAnnotationResponse(ids=ids)


async def list_by(
    db: Session,
    task_id: Optional[int],
    sample_name: Optional[str],
    after: Optional[int],
    before: Optional[int],
    pageNo: Optional[int],
    pageSize: int,
    sorting: Optional[str],
    current_user: User,
) -> Tuple[List[PreAnnotationResponse], int]:
    
    pre_annotations = []
    
    # 指定 sample_name 查询时，需要对所有的 pre_annotation 进行查询
    # TODO 优化查询逻辑：将上传的jsonl中的每个预标注提取并存储到单独的表（PreAnnotationDetail）中
    if sample_name:
        pre_annotations = db.query(TaskPreAnnotation).all()
    else:
        pre_annotations = crud_pre_annotation.list_by(
            db=db,
            task_id=task_id,
            owner_id=current_user.id,
            after=after,
            before=before,
            pageNo=pageNo,
            pageSize=None if sample_name else pageSize,
            sorting=sorting,
        )


    total = crud_pre_annotation.count(db=db, task_id=task_id, owner_id=current_user.id)

    def parse_jsonl_file(file_id: int, sample_name: str) -> List[dict]:
        jsonl_content = read_jsonl_file(db, file_id)
        # Filter by sample_name
        result = []
        for item in jsonl_content:
            if sample_name is None or item.get("sample_name") == sample_name:
                result.append(item)
        return result
    
    real_sample_name = sample_name[9:] if sample_name else None
    parsed_data_list = [parse_jsonl_file(pre_annotation.file_id, real_sample_name) for pre_annotation in pre_annotations]

    def is_contains_sample_name(inputs: List[dict], sample_name: str) -> bool:
        return any(data.get("sample_name") == sample_name for data in inputs)

    filtered_pre_annotations = []
    for i, pre_annotation in enumerate(pre_annotations):
        if real_sample_name is None or is_contains_sample_name(parsed_data_list[i], real_sample_name):
            filtered_pre_annotations.append(
                PreAnnotationResponse(
                    id=pre_annotation.id,
                    file=AttachmentResponse(id=pre_annotation.file.id, filename=pre_annotation.file.filename, url=pre_annotation.file.url) if pre_annotation.file else None,
                    created_at=pre_annotation.created_at,
                    data=parsed_data_list[i],
                    created_by=UserResp(
                        id=pre_annotation.owner.id,
                        username=pre_annotation.owner.username,
                    ),
                    updated_at=pre_annotation.updated_at,
                    updated_by=UserResp(
                        id=pre_annotation.updater.id,
                        username=pre_annotation.updater.username,
                    ),
                )
            )
    
    return filtered_pre_annotations, total


async def get(
    db: Session, task_id: int, pre_annotation_id: int, current_user: User
) -> PreAnnotationResponse:
    pre_annotation = crud_pre_annotation.get(
        db=db,
        pre_annotation_id=pre_annotation_id,
    )

    if not pre_annotation:
        logger.error("cannot find pre_annotation: {}", pre_annotation_id)
        raise LabelUException(
            code=ErrorCode.CODE_55001_SAMPLE_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )
        
    # response
    return PreAnnotationResponse(
        id=pre_annotation.id,
        file=AttachmentResponse(id=pre_annotation.file.id, filename=pre_annotation.file.filename, url=pre_annotation.file.url) if pre_annotation.file else None,
        created_at=pre_annotation.created_at,
        created_by=UserResp(
            id=pre_annotation.owner.id,
            username=pre_annotation.owner.username,
        ),
        updated_at=pre_annotation.updated_at,
        updated_by=UserResp(
            id=pre_annotation.updater.id,
            username=pre_annotation.updater.username,
        ),
    )

async def delete(
    db: Session, pre_annotation_ids: List[int], current_user: User
) -> CommonDataResp:

    with db.begin():
        crud_pre_annotation.delete(db=db, pre_annotation_ids=pre_annotation_ids)
    # response
    return CommonDataResp(ok=True)



from datetime import datetime

from sqlalchemy.schema import Index
from sqlalchemy.orm import relationship
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Text

from labelu.internal.common.db import Base


class TaskPreAnnotation(Base):
    __tablename__ = "task_pre_annotation"

    id = Column(Integer, primary_key=True, autoincrement=True, index=True)
    task_id = Column(Integer, ForeignKey("task.id"), index=True)
    file_id = Column(Integer, ForeignKey("task_attachment.id"), index=True)
    created_by = Column(Integer, ForeignKey("user.id"), index=True)
    updated_by = Column(Integer, ForeignKey("user.id"))
    created_at = Column(
        DateTime, default=datetime.now, comment="Time a task sample result was created"
    )
    updated_at = Column(
        DateTime,
        default=datetime.now,
        onupdate=datetime.now,
        comment="Last time a task pre annotation result was updated",
    )
    data = Column(Text, comment="task sample pre annotation result")
    deleted_at = Column(DateTime, index=True, comment="Task delete time")

    file = relationship("TaskAttachment", foreign_keys=[file_id])
    task = relationship("Task", foreign_keys=[task_id])
    owner = relationship("User", foreign_keys=[created_by])
    updater = relationship("User", foreign_keys=[updated_by])

    Index("idx_pre_annotation_id_deleted_at", id, deleted_at)




import aiofiles
import os
import uuid
from PIL import Image
from pathlib import Path

from loguru import logger
from fastapi import status
from sqlalchemy.orm import Session

from labelu.internal.common.config import settings
from labelu.internal.common.error_code import ErrorCode
from labelu.internal.common.error_code import LabelUException
from labelu.internal.domain.models.user import User
from labelu.internal.domain.models.task import TaskStatus
from labelu.internal.domain.models.attachment import TaskAttachment
from labelu.internal.adapter.persistence import crud_task
from labelu.internal.adapter.persistence import crud_attachment
from labelu.internal.application.command.attachment import AttachmentCommand
from labelu.internal.application.command.attachment import AttachmentDeleteCommand
from labelu.internal.application.response.base import CommonDataResp
from labelu.internal.application.response.attachment import AttachmentResponse


async def create(
    db: Session, task_id: int, cmd: AttachmentCommand, current_user: User
) -> AttachmentResponse:

    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task: {}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_50002_TASK_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    # save file
    try:
        # file relative path
        path_filename = cmd.file.filename.split("/")
        filename = str(uuid.uuid4())[0:8] + "-" + path_filename[-1]
        path = "/".join(path_filename[:-1])
        attachment_relative_base_dir = Path(settings.UPLOAD_DIR).joinpath(
            str(task_id), path
        )
        attachment_relative_path = str(attachment_relative_base_dir.joinpath(filename))

        # file full path
        attachment_full_base_dir = Path(settings.MEDIA_ROOT).joinpath(
            attachment_relative_base_dir
        )
        attachment_full_path = Path(settings.MEDIA_ROOT).joinpath(
            attachment_relative_path
        )

        # create dicreatory
        attachment_full_base_dir.mkdir(parents=True, exist_ok=True)

        # save image
        logger.info(attachment_full_path)
        async with aiofiles.open(attachment_full_path, "wb") as out_file:
            content = await cmd.file.read()  # async read
            await out_file.write(content)  # async write

        # create thumbnail for image
        if cmd.file.content_type.startswith("image/"):
            tumbnail_full_path = Path(
                f"{attachment_full_path.parent}/{attachment_full_path.stem}-thumbnail{attachment_full_path.suffix}"
            )
            logger.info(tumbnail_full_path)
            image = Image.open(attachment_full_path)
            image.thumbnail(
                (
                    round(image.width / image.height * settings.THUMBNAIL_HEIGH_PIXEL),
                    settings.THUMBNAIL_HEIGH_PIXEL,
                ),
            )
            if image.mode != "RGB":
                image = image.convert("RGB")
            image.save(tumbnail_full_path)
    except Exception as e:
        logger.error(e)
        raise LabelUException(
            code=ErrorCode.CODE_51000_CREATE_ATTACHMENT_ERROR,
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )

    # check file already saved
    if not attachment_full_path.exists() or (
        cmd.file.content_type.startswith("image/") and not tumbnail_full_path.exists()
    ):
        logger.error(
            "cannot find saved images, path is:{}, image content-type is:{}, thumbnail path is:{}",
            attachment_full_path,
            cmd.file.content_type,
            tumbnail_full_path,
        )
        raise LabelUException(
            code=ErrorCode.CODE_51000_CREATE_ATTACHMENT_ERROR,
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )

    attachment_url_path = attachment_relative_path.replace("\\", "/")
    attachment_api_url = f"{settings.API_V1_STR}/tasks/attachment/{attachment_url_path}"
    # add a task file record
    with db.begin():
        attachment = crud_attachment.create(
            db=db,
            attachment=TaskAttachment(
                path=attachment_url_path,
                url=attachment_api_url,
                filename=filename,
                created_by=current_user.id,
                updated_by=current_user.id,
                task_id=task_id,
            ),
        )

    # response
    return AttachmentResponse(
        id=attachment.id,
        url=attachment_api_url,
        filename=filename,
    )


async def download_attachment(file_path: str) -> str:

    # check file exist
    file_full_path = settings.MEDIA_ROOT.joinpath(file_path.lstrip("/"))
    if not file_full_path.is_file() or not file_full_path.exists():
        logger.error("attachment not found:{}", file_full_path)
        raise LabelUException(
            code=ErrorCode.CODE_51001_TASK_ATTACHMENT_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    # response
    return file_full_path


async def delete(
    db: Session, task_id: int, cmd: AttachmentDeleteCommand, current_user: User
) -> CommonDataResp:

    # get task
    task = crud_task.get(db=db, task_id=task_id)
    if not task:
        logger.error("cannot find task:{}", task_id)
        raise LabelUException(
            code=ErrorCode.CODE_51001_TASK_ATTACHMENT_NOT_FOUND,
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if task.created_by != current_user.id:
        logger.error(
            "cannot delete attachment, the task owner is:{}, the delete operator is:{}",
            task.created_by,
            current_user.id,
        )
        raise LabelUException(
            code=ErrorCode.CODE_30001_NO_PERMISSION,
            status_code=status.HTTP_403_FORBIDDEN,
        )

    # delete media
    try:
        attachments = crud_attachment.get_by_ids(
            db=db, attachment_ids=cmd.attachment_ids
        )
        for attachment in attachments:
            file_full_path = Path(settings.MEDIA_ROOT).joinpath(attachment.path)
            os.remove(file_full_path)
    except Exception as e:
        logger.error(e)

    # delete
    with db.begin():
        crud_attachment.delete(db=db, attachment_ids=cmd.attachment_ids)

    # response
    return CommonDataResp(ok=True)



import json

from alembic import op
from sqlalchemy import engine_from_config
from sqlalchemy.engine import reflection

def table_exist(table_name):
    """check table is not exist

    Args:
        table_name (string): the name of table

    Returns:
        bool: true or false, whether the table_name exists
    """
    config = op.get_context().config
    engine = engine_from_config(
        config.get_section(config.config_ini_section), prefix="sqlalchemy."
    )
    insp = reflection.Inspector.from_engine(engine)
    table_exist = False
    
    for table in insp.get_table_names():
        if table_name not in table:
            continue
        table_exist = True
        break
    
    return table_exist

def column_exist_in_table(table_name, column_name):
    """check column is not exist in table

    Args:
        table_name (string): the name of table
        column_name (string): the name of column

    Returns:
        bool: true or false, whether the column_name exists in the table_name
    """
    config = op.get_context().config
    engine = engine_from_config(
        config.get_section(config.config_ini_section), prefix="sqlalchemy."
    )
    insp = reflection.Inspector.from_engine(engine)
    column_exist = False
    for col in insp.get_columns(table_name):
        if column_name not in col["name"]:
            continue
        column_exist = True
    return column_exist


def get_tool_label_dict(task_config: dict) -> dict:
    """get the key value of labels in a given task_id and task_config"""

    label_dict = {"无标签": "noneAttribute"}
    # obtain the labels in current task
    # get the general labels
    for normal_label in task_config.get("attribute", []):
        if normal_label.get("key", ""):
            label_dict[normal_label.get("key")] = normal_label.get("value")
    # get the labels in configuration defined by user
    for task_tool in task_config.get("tools", []):
        if "config" not in task_tool.keys():
            continue
        labels = task_tool.get("config").get("attributeList", [])
        for label in labels:
            if label.get("key", ""):
                label_dict[label.get("key")] = label.get("value")
    return label_dict


def replace_key_with_value(sample_data: dict, label_dict: dict) -> dict:
    """replace the key with value in task_sample table to modify the error for the history version"""

    annotated_result = sample_data.get("result")
    annotated_result = json.loads(annotated_result)
    for sample_tool, sample_tool_results in annotated_result.items():
        if sample_tool.endswith("Tool"):
            for sample_tool_result in sample_tool_results.get("result", []):
                tool_label = sample_tool_result.get("attribute", "")
                if tool_label in label_dict:
                    sample_tool_result["attribute"] = label_dict[tool_label]
    return annotated_result



import os
from pathlib import Path

from loguru import logger
from pydantic import BaseSettings

from labelu.internal.common.io import get_data_dir


class Settings(BaseSettings):
    SCHEME: str = "http"
    HOST: str = "localhost"
    PORT: str = "8000"
    API_V1_STR: str = "/api/v1"
    MEDIA_HOST: str = f"{SCHEME}://{HOST}:{PORT}"

    BASE_DATA_DIR = get_data_dir()
    MEDIA_ROOT = Path(BASE_DATA_DIR).joinpath("media")
    UPLOAD_DIR = "upload"
    EXPORT_DIR = "export"
    os.makedirs(MEDIA_ROOT, exist_ok=True)
    logger.info("Database and media directory: {}", BASE_DATA_DIR)
    UPLOAD_FILE_MAX_SIZE = 200_000_000  # ~200MB
    THUMBNAIL_HEIGH_PIXEL = 120

    DATABASE_URL: str = f"sqlite:///{BASE_DATA_DIR}/labelu.sqlite"

    PASSWORD_SECRET_KEY = (
        "e5b7d00a59aaa2a5ea86a7c4d72f856b20bafa1b8d0e66124082ada81f6340bd"
    )

    TOKEN_GENERATE_ALGORITHM = "HS256"
    TOKEN_ACCESS_EXPIRE_MINUTES = 30
    TOKEN_TYPE = "Bearer"

    class Config:
        env_prefix = ""
        env_file_encoding = "utf-8"
        case_sensitive = True


settings = Settings()



from datetime import datetime
import json
from typing import Any, Dict, List, Union


from sqlalchemy import case, text
from sqlalchemy.orm import Session
from fastapi.encoders import jsonable_encoder

from labelu.internal.domain.models.pre_annotation import TaskPreAnnotation
from labelu.internal.adapter.persistence import crud_attachment


def batch(db: Session, pre_annotations: List[TaskPreAnnotation]) -> List[TaskPreAnnotation]:
    db.bulk_save_objects(pre_annotations, return_defaults=True)
    return pre_annotations


def list_by(
    db: Session,
    task_id: Union[int, None],
    owner_id: int,
    after: Union[int, None],
    before: Union[int, None],
    pageNo: Union[int, None],
    pageSize: int,
    sorting: Union[str, None],
) -> List[TaskPreAnnotation]:

    # query filter
    query_filter = [TaskPreAnnotation.created_by == owner_id, TaskPreAnnotation.deleted_at == None]
    if before:
        query_filter.append(TaskPreAnnotation.id < before)
    if after:
        query_filter.append(TaskPreAnnotation.id > after)
    if task_id:
        query_filter.append(TaskPreAnnotation.task_id == task_id)
    query = db.query(TaskPreAnnotation).filter(*query_filter)

    # default order by id, before need select last items
    if before:
        query = query.order_by(TaskPreAnnotation.id.desc())
    else:
        query = query.order_by(TaskPreAnnotation.id.asc())
    results = (
        query.offset(offset=pageNo * pageSize if pageNo else 0)
        .limit(limit=pageSize)
        .all()
    )
    
    # No sorting
    
    if before:
        results.reverse()
        
    return results

def list_by_task_id_and_owner_id(db: Session, task_id: int, owner_id: int) -> Dict[str, List[TaskPreAnnotation]]:
    pre_annotations = db.query(TaskPreAnnotation).filter(
        TaskPreAnnotation.task_id == task_id,
        TaskPreAnnotation.deleted_at == None,
        TaskPreAnnotation.created_by == owner_id
    ).all()
    
    return pre_annotations

def get(db: Session, pre_annotation_id: int) -> TaskPreAnnotation:
    return (
        db.query(TaskPreAnnotation)
        .filter(TaskPreAnnotation.id == pre_annotation_id, TaskPreAnnotation.deleted_at == None)
        .first()
    )


def get_by_ids(db: Session, pre_annotation_ids: List[int]) -> List[TaskPreAnnotation]:
    return (
        db.query(TaskPreAnnotation)
        .filter(TaskPreAnnotation.id.in_(pre_annotation_ids), TaskPreAnnotation.deleted_at == None)
        .all()
    )


def update(db: Session, db_obj: TaskPreAnnotation, obj_in: Dict[str, Any]) -> TaskPreAnnotation:
    obj_data = jsonable_encoder(obj_in)
    for field in obj_data:
        if field in obj_in:
            setattr(db_obj, field, obj_in[field])
    db.add(db_obj)
    db.flush()
    db.refresh(db_obj)
    return db_obj


def delete(db: Session, pre_annotation_ids: List[int]) -> None:
    db.query(TaskPreAnnotation).filter(TaskPreAnnotation.id.in_(pre_annotation_ids)).update(
        {TaskPreAnnotation.deleted_at: datetime.now()}
    )


def count(db: Session, task_id: int, owner_id: int) -> int:
    query_filter = [TaskPreAnnotation.created_by == owner_id, TaskPreAnnotation.deleted_at == None]
    if task_id:
        query_filter.append(TaskPreAnnotation.task_id == task_id)
    return db.query(TaskPreAnnotation).filter(*query_filter).count()



import os
import sys
from logging.config import fileConfig

from sqlalchemy import engine_from_config
from sqlalchemy import pool

from alembic import context

from labelu.internal.common.db import Base
from labelu.internal.domain.models.task import Task
from labelu.internal.domain.models.user import User
from labelu.internal.domain.models.sample import TaskSample
from labelu.internal.domain.models.attachment import TaskAttachment
from labelu.internal.common.config import settings


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section, {}),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from enum import Enum
from datetime import datetime

from sqlalchemy.schema import Index
from sqlalchemy.orm import relationship
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Text

from labelu.internal.common.db import Base


class TaskStatus(str, Enum):
    """
    task status
    """

    DRAFT = "DRAFT"
    IMPORTED = "IMPORTED"
    CONFIGURED = "CONFIGURED"
    INPROGRESS = "INPROGRESS"
    FINISHED = "FINISHED"


class MediaType(str, Enum):
    """
    task meida type
    """

    IMAGE = "IMAGE"
    VIDEO = "VIDEO"
    AUDIO = "AUDIO"


class Task(Base):
    __tablename__ = "task"

    id = Column(Integer, primary_key=True, autoincrement=True, index=True)
    name = Column(String(64), index=True)
    description = Column(String(1024), comment="task description")
    tips = Column(String(1024), comment="task tips")
    last_sample_inner_id = Column(
        Integer, default=0, comment="The last inner id of sample in a task"
    )
    config = Column(Text, comment="task config yaml")
    media_type = Column(
        String(32),
        comment="task media type: image, video, audio",
    )
    status = Column(String(32), default=TaskStatus.DRAFT.value, comment="task status")
    created_by = Column(Integer, ForeignKey(column="user.id"), index=True)
    updated_by = Column(
        Integer, ForeignKey(column="user.id"), comment="Last time a task was updated by"
    )
    created_at = Column(
        DateTime, default=datetime.now, comment="Time a task was created"
    )
    updated_at = Column(
        DateTime,
        default=datetime.now,
        onupdate=datetime.now,
        comment="Last time a task was updated",
    )
    deleted_at = Column(DateTime, index=True, comment="Task delete time")

    owner = relationship("User", foreign_keys=[created_by])
    updater = relationship("User", foreign_keys=[updated_by])

    Index("idx_task_id_deleted_at_index", id, deleted_at)



from datetime import datetime
from typing import Any, Dict, List

from sqlalchemy.orm import Session
from fastapi.encoders import jsonable_encoder

from labelu.internal.domain.models.task import Task


def create(db: Session, task: Task) -> Task:
    db.add(task)
    db.flush()
    db.refresh(task)
    return task


def list_by(db: Session, owner_id: int, page: int = 0, size: int = 100) -> List[Task]:
    return (
        db.query(Task)
        .filter(Task.created_by == owner_id, Task.deleted_at == None)
        .order_by(Task.id.desc())
        .offset(offset=page * size)
        .limit(limit=size)
        .all()
    )


def get(db: Session, task_id: int, lock: bool = False) -> Task:
    """return a Task item according to task id

    Args:
        db (Session): a session connect to table
        task_id (int): a given task id
        lock (bool, optional): if lock is True, means add exclusive lock in a row. Defaults to False.
    """
    if not lock:
        return (
            db.query(Task).filter(Task.id == task_id, Task.deleted_at == None).first()
        )
    else:
        return (
            db.query(Task)
            .filter(Task.id == task_id, Task.deleted_at == None)
            .with_for_update()
            .first()
        )


def update(db: Session, db_obj: Task, obj_in: Dict[str, Any]) -> Task:
    obj_data = jsonable_encoder(obj_in)
    for field in obj_data:
        if field in obj_in:
            setattr(db_obj, field, obj_in[field])
    db.add(db_obj)
    db.flush()
    db.refresh(db_obj)
    return db_obj


def delete(db: Session, db_obj: Task) -> Task:
    db_obj.deleted_at = datetime.now()
    db.add(db_obj)
    db.flush()
    db.refresh(db_obj)
    return db_obj


def count(db: Session, owner_id: int) -> int:
    return (
        db.query(Task)
        .filter(Task.created_by == owner_id, Task.deleted_at == None)
        .count()
    )



from typing import List

from sqlalchemy.orm import Session
from labelu.internal.domain.models.user import User


def create(db: Session, user: User) -> User:
    db_user = User(username=user.username, hashed_password=user.hashed_password)
    db.add(db_user)
    db.flush()
    db.refresh(db_user)
    return db_user


def get(db: Session, id: int) -> User:
    return db.query(User).filter(User.id == id).first()


def get_user_by_username(db: Session, username: str) -> User:
    return db.query(User).filter(User.username == username).first()



from datetime import datetime
from typing import List

from sqlalchemy.orm import Session

from labelu.internal.domain.models.attachment import TaskAttachment


def create(db: Session, attachment: TaskAttachment) -> TaskAttachment:
    db.add(attachment)
    db.flush()
    db.refresh(attachment)
    return attachment


def get(db: Session, attachment_id: int) -> TaskAttachment:
    return (
        db.query(TaskAttachment)
        .filter(TaskAttachment.id == attachment_id, TaskAttachment.deleted_at == None)
        .first()
    )


def get_by_ids(db: Session, attachment_ids: List[int]) -> List[TaskAttachment]:
    return (
        db.query(TaskAttachment)
        .filter(
            TaskAttachment.id.in_(attachment_ids), TaskAttachment.deleted_at == None
        )
        .all()
    )


def delete(db: Session, attachment_ids: List[int]) -> None:
    db.query(TaskAttachment).filter(TaskAttachment.id.in_(attachment_ids)).update(
        {TaskAttachment.deleted_at: datetime.now()}
    )



from enum import Enum
from typing import List, Union
from pydantic import BaseModel, Field

from labelu.internal.domain.models.sample import SampleState


class ExportType(str, Enum):
    """
    export samle result file type
    """

    JSON = "JSON"
    MASK = "MASK"
    COCO = "COCO"


class CreateSampleCommand(BaseModel):
    file_id: int = Field(
        gt=0,
        description="description: attachment file id",
    )
    data: Union[dict, None] = Field(
        default=None,
        description="description: annotation result of sample",
    )


class DeleteSampleCommand(BaseModel):
    sample_ids: List[int] = Field(
        min_items=1,
        description="description: attachment file id",
    )


class PatchSampleCommand(BaseModel):
    data: Union[dict, None] = Field(
        default=None,
        description="description: sample data, include result",
    )
    annotated_count: Union[int, None] = Field(
        default=0, description="description: annotate result count"
    )
    state: Union[SampleState, None] = Field(
        description="description: sample file state, must be 'SKIPPED', 'NEW', or None",
    )


class ExportSampleCommand(BaseModel):
    sample_ids: Union[List[int], None] = Field(
        min_items=1,
        gt=0,
        description="description: sample id",
    )



from sqlalchemy.orm import Session
from pathlib import Path
from fastapi.testclient import TestClient

from labelu.internal.common.config import settings
from labelu.internal.adapter.persistence import crud_user
from labelu.internal.adapter.persistence import crud_task
from labelu.internal.adapter.persistence import crud_pre_annotation
from labelu.internal.adapter.persistence import crud_sample
from labelu.internal.adapter.persistence import crud_attachment
from labelu.internal.domain.models.task import Task
from labelu.internal.domain.models.attachment import TaskAttachment
from labelu.internal.domain.models.pre_annotation import TaskPreAnnotation
from labelu.internal.domain.models.sample import TaskSample


class TestClassTaskPreAnnotationRouter:
    def test_create_pre_annotation_successful(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:
            
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )
        
        # prepare data
        with Path("labelu/tests/data/test.jsonl").open(mode="rb") as f:
            jsonl = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
            
        data = [{"file_id": jsonl.json()["data"]["id"]}]

        # run
        r = client.post(
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations",
            headers=testuser_token_headers,
            json=data,
        )

        json = r.json()
        assert r.status_code == 201
        assert len(json["data"]["ids"]) == 1
        
    def test_create_pre_annotation_sample_exists(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:
            
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )
        
        # prepare data
        with Path("labelu/tests/data/test.jsonl").open(mode="rb") as f:
            res = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
            
        data = [{"file_id": res.json()["data"]["id"]}]

        # run
        r = client.post(
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations",
            headers=testuser_token_headers,
            json=data,
        )
        
        with Path("labelu/tests/data/test.jsonl").open(mode="rb") as f:
            res = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
            
        data = [{"file_id": res.json()["data"]["id"]}]

        # run
        r = client.post(
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations",
            headers=testuser_token_headers,
            json=data,
        )

        assert r.status_code == 400

    def test_create_sample_task_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = [{"file_id": 1, "data": {}}]

        # run
        r = client.post(
            f"{settings.API_V1_STR}/tasks/0/pre_annotations",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        json = r.json()
        assert r.status_code == 404
        assert json["err_code"] == 50002

    def test_pre_annotation_list_by_page(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )
        
        with Path("labelu/tests/data/test.jsonl").open(mode="rb") as f:
            jsonl = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        
        pre_annotations = [
            TaskPreAnnotation(
                task_id=task.id,
                file_id=jsonl.json()["data"]["id"],
                created_by=current_user.id,
                updated_by=current_user.id,
            )
            for i in range(14)
        ]
        
        crud_pre_annotation.batch(
            db=db,
            pre_annotations=pre_annotations,
        )

        # run
        r = client.get(
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations",
            headers=testuser_token_headers,
            params={"pageNo": 0, "pageSize": 10},
        )

        # check
        json = r.json()
        assert r.status_code == 200
        assert len(json["data"]) == 10
        assert json["data"][0]["id"] == 1
        assert json["meta_data"]["total"] == 14

    def test_pre_annotation_list_with_sample_name(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )
        
        # upload sample file
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            img = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        # upload jsonl file
        with Path("labelu/tests/data/test.jsonl").open(mode="rb") as f:
            jsonl = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        
        samples = [
            TaskSample(
                task_id=task.id,
                file_id=img.json()["data"]["id"],
                created_by=current_user.id,
                updated_by=current_user.id,
                data="{}",
            )
        ]
        crud_sample.batch(
            db=db,
            samples=samples,
        )
        
        pre_annotations = [
            TaskPreAnnotation(
                task_id=task.id,
                file_id=jsonl.json()["data"]["id"],
                created_by=current_user.id,
                updated_by=current_user.id,
            )
        ]
        crud_pre_annotation.batch(
            db=db,
            pre_annotations=pre_annotations,
        )

        # run
        r = client.get(
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations?sample_name={img.json()['data']['filename']}",
            headers=testuser_token_headers,
            params={"pageNo": 0},
        )

        # check
        json = r.json()
        assert r.status_code == 200
        assert json["data"][0]["id"] == 1
        assert json["meta_data"]["total"] == 1

    def test_pre_annotation_list_with_sample_name_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )
        
        # upload sample file
        with Path("labelu/tests/data/test.png").open(mode="rb") as f:
            img = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        # upload jsonl file
        with Path("labelu/tests/data/test.jsonl").open(mode="rb") as f:
            jsonl = client.post(
                f"{settings.API_V1_STR}/tasks/{task.id}/attachments",
                headers=testuser_token_headers,
                files={"file": f},
            )
        
        samples = [
            TaskSample(
                task_id=task.id,
                file_id=img.json()["data"]["id"],
                created_by=current_user.id,
                updated_by=current_user.id,
                data="{}",
            )
        ]
        crud_sample.batch(
            db=db,
            samples=samples,
        )
        
        pre_annotations = [
            TaskPreAnnotation(
                task_id=task.id,
                file_id=jsonl.json()["data"]["id"],
                created_by=current_user.id,
                updated_by=current_user.id,
            )
        ]
        crud_pre_annotation.batch(
            db=db,
            pre_annotations=pre_annotations,
        )

        # run
        r = client.get(
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations?sample_name=asdasd.png",
            headers=testuser_token_headers,
            params={"pageNo": 0},
        )

        # check
        json = r.json()
        assert r.status_code == 200
        assert len(json["data"]) == 0
    
    def test_pre_annotations_list_by_params_error(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data

        # run
        r = client.get(
            f"{settings.API_V1_STR}/tasks/{1}/pre_annotations",
            headers=testuser_token_headers,
            params={"after": 1, "before": 1, "pageNo": 1, "pageSize": 10},
        )

        # check
        assert r.status_code == 422

    def test_sample_list_by_sample_name_error(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )
        # run
        r = client.get(
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations?sample_name=",
            headers=testuser_token_headers,
            params={"pageNo": 0, "pageSize": 10},
        )

        # check
        assert r.status_code == 422

    def test_pre_annotation_get(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )
        pre_annotations = crud_pre_annotation.batch(
            db=db,
            pre_annotations=[
                TaskPreAnnotation(
                    task_id=task.id,
                    file_id=1,
                    created_by=current_user.id,
                    updated_by=current_user.id,
                )
            ],
        )
        
        assert pre_annotations[0] is not None

        # run
        r = client.get(
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations/{pre_annotations[0].id}",
            headers=testuser_token_headers,
        )

        # check
        assert r.status_code == 200

    def test_pre_annotation_delete(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )
        pre_annotations = crud_pre_annotation.batch(
            db=db,
            pre_annotations=[
                TaskPreAnnotation(
                    task_id=1,
                    file_id=1,
                    created_by=current_user.id,
                    updated_by=current_user.id,
                    data="{}",
                )
            ],
        )

        # run
        data = {"pre_annotation_ids": [pre_annotations[0].id]}
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        assert r.status_code == 200

    def test_pre_annotations_delete_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )

        # run
        data = {"pre_annotation_ids": [1, 2]}
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/{task.id}/pre_annotations",
            headers=testuser_token_headers,
            json=data,
        )

        # check
        assert r.status_code == 200
